/****************************************************************************
** Meta object code from reading C++ file 'theme.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.1)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../syntax-highlighting-6.27.0/src/lib/theme.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'theme.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.1. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN19KSyntaxHighlighting5ThemeE_t {};
} // unnamed namespace

template <> constexpr inline auto KSyntaxHighlighting::Theme::qt_create_metaobjectdata<qt_meta_tag_ZN19KSyntaxHighlighting5ThemeE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "KSyntaxHighlighting::Theme",
        "name",
        "translatedName",
        "TextStyle",
        "Normal",
        "Keyword",
        "Function",
        "Variable",
        "ControlFlow",
        "Operator",
        "BuiltIn",
        "Extension",
        "Preprocessor",
        "Attribute",
        "Char",
        "SpecialChar",
        "String",
        "VerbatimString",
        "SpecialString",
        "Import",
        "DataType",
        "DecVal",
        "BaseN",
        "Float",
        "Constant",
        "Comment",
        "Documentation",
        "Annotation",
        "CommentVar",
        "RegionMarker",
        "Information",
        "Warning",
        "Alert",
        "Error",
        "Others",
        "EditorColorRole",
        "BackgroundColor",
        "TextSelection",
        "CurrentLine",
        "SearchHighlight",
        "ReplaceHighlight",
        "BracketMatching",
        "TabMarker",
        "SpellChecking",
        "IndentationLine",
        "IconBorder",
        "CodeFolding",
        "LineNumbers",
        "CurrentLineNumber",
        "WordWrapMarker",
        "ModifiedLines",
        "SavedLines",
        "Separator",
        "MarkBookmark",
        "MarkBreakpointActive",
        "MarkBreakpointReached",
        "MarkBreakpointDisabled",
        "MarkExecution",
        "MarkWarning",
        "MarkError",
        "TemplateBackground",
        "TemplatePlaceholder",
        "TemplateFocusedPlaceholder",
        "TemplateReadOnlyPlaceholder"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
        // property 'name'
        QtMocHelpers::PropertyData<QString>(1, QMetaType::QString, QMC::DefaultPropertyFlags),
        // property 'translatedName'
        QtMocHelpers::PropertyData<QString>(2, QMetaType::QString, QMC::DefaultPropertyFlags),
    };
    QtMocHelpers::UintData qt_enums {
        // enum 'TextStyle'
        QtMocHelpers::EnumData<enum TextStyle>(3, 3, QMC::EnumFlags{}).add({
            {    4, TextStyle::Normal },
            {    5, TextStyle::Keyword },
            {    6, TextStyle::Function },
            {    7, TextStyle::Variable },
            {    8, TextStyle::ControlFlow },
            {    9, TextStyle::Operator },
            {   10, TextStyle::BuiltIn },
            {   11, TextStyle::Extension },
            {   12, TextStyle::Preprocessor },
            {   13, TextStyle::Attribute },
            {   14, TextStyle::Char },
            {   15, TextStyle::SpecialChar },
            {   16, TextStyle::String },
            {   17, TextStyle::VerbatimString },
            {   18, TextStyle::SpecialString },
            {   19, TextStyle::Import },
            {   20, TextStyle::DataType },
            {   21, TextStyle::DecVal },
            {   22, TextStyle::BaseN },
            {   23, TextStyle::Float },
            {   24, TextStyle::Constant },
            {   25, TextStyle::Comment },
            {   26, TextStyle::Documentation },
            {   27, TextStyle::Annotation },
            {   28, TextStyle::CommentVar },
            {   29, TextStyle::RegionMarker },
            {   30, TextStyle::Information },
            {   31, TextStyle::Warning },
            {   32, TextStyle::Alert },
            {   33, TextStyle::Error },
            {   34, TextStyle::Others },
        }),
        // enum 'EditorColorRole'
        QtMocHelpers::EnumData<enum EditorColorRole>(35, 35, QMC::EnumFlags{}).add({
            {   36, EditorColorRole::BackgroundColor },
            {   37, EditorColorRole::TextSelection },
            {   38, EditorColorRole::CurrentLine },
            {   39, EditorColorRole::SearchHighlight },
            {   40, EditorColorRole::ReplaceHighlight },
            {   41, EditorColorRole::BracketMatching },
            {   42, EditorColorRole::TabMarker },
            {   43, EditorColorRole::SpellChecking },
            {   44, EditorColorRole::IndentationLine },
            {   45, EditorColorRole::IconBorder },
            {   46, EditorColorRole::CodeFolding },
            {   47, EditorColorRole::LineNumbers },
            {   48, EditorColorRole::CurrentLineNumber },
            {   49, EditorColorRole::WordWrapMarker },
            {   50, EditorColorRole::ModifiedLines },
            {   51, EditorColorRole::SavedLines },
            {   52, EditorColorRole::Separator },
            {   53, EditorColorRole::MarkBookmark },
            {   54, EditorColorRole::MarkBreakpointActive },
            {   55, EditorColorRole::MarkBreakpointReached },
            {   56, EditorColorRole::MarkBreakpointDisabled },
            {   57, EditorColorRole::MarkExecution },
            {   58, EditorColorRole::MarkWarning },
            {   59, EditorColorRole::MarkError },
            {   60, EditorColorRole::TemplateBackground },
            {   61, EditorColorRole::TemplatePlaceholder },
            {   62, EditorColorRole::TemplateFocusedPlaceholder },
            {   63, EditorColorRole::TemplateReadOnlyPlaceholder },
        }),
    };
    return QtMocHelpers::metaObjectData<Theme, qt_meta_tag_ZN19KSyntaxHighlighting5ThemeE_t>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject KSyntaxHighlighting::Theme::staticMetaObject = { {
    nullptr,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19KSyntaxHighlighting5ThemeE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19KSyntaxHighlighting5ThemeE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN19KSyntaxHighlighting5ThemeE_t>.metaTypes,
    nullptr
} };

void KSyntaxHighlighting::Theme::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = reinterpret_cast<Theme *>(_o);
    if (_c == QMetaObject::ReadProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: *reinterpret_cast<QString*>(_v) = _t->name(); break;
        case 1: *reinterpret_cast<QString*>(_v) = _t->translatedName(); break;
        default: break;
        }
    }
}
QT_WARNING_POP
