/****************************************************************************
** Meta object code from reading C++ file 'repository.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.1)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../syntax-highlighting-6.27.0/src/lib/repository.h"
#include <QtCore/qmetatype.h>
#include <QtCore/QList>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'repository.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.1. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN19KSyntaxHighlighting10RepositoryE_t {};
} // unnamed namespace

template <> constexpr inline auto KSyntaxHighlighting::Repository::qt_create_metaobjectdata<qt_meta_tag_ZN19KSyntaxHighlighting10RepositoryE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "KSyntaxHighlighting::Repository",
        "aboutToReload",
        "",
        "reloaded",
        "definitionForName",
        "KSyntaxHighlighting::Definition",
        "defName",
        "definitionForFileName",
        "fileName",
        "definitionsForFileName",
        "QList<KSyntaxHighlighting::Definition>",
        "definitionForMimeType",
        "mimeType",
        "definitionsForMimeType",
        "definitions",
        "themes",
        "QList<KSyntaxHighlighting::Theme>",
        "theme",
        "KSyntaxHighlighting::Theme",
        "themeName",
        "defaultTheme",
        "KSyntaxHighlighting::Repository::DefaultTheme",
        "t",
        "DefaultTheme",
        "LightTheme",
        "DarkTheme"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'aboutToReload'
        QtMocHelpers::SignalData<void()>(1, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'reloaded'
        QtMocHelpers::SignalData<void()>(3, 2, QMC::AccessPublic, QMetaType::Void),
        // Method 'definitionForName'
        QtMocHelpers::MethodData<KSyntaxHighlighting::Definition(const QString &) const>(4, 2, QMC::AccessPublic, 0x80000000 | 5, {{
            { QMetaType::QString, 6 },
        }}),
        // Method 'definitionForFileName'
        QtMocHelpers::MethodData<KSyntaxHighlighting::Definition(const QString &) const>(7, 2, QMC::AccessPublic, 0x80000000 | 5, {{
            { QMetaType::QString, 8 },
        }}),
        // Method 'definitionsForFileName'
        QtMocHelpers::MethodData<QList<KSyntaxHighlighting::Definition>(const QString &) const>(9, 2, QMC::AccessPublic, 0x80000000 | 10, {{
            { QMetaType::QString, 8 },
        }}),
        // Method 'definitionForMimeType'
        QtMocHelpers::MethodData<KSyntaxHighlighting::Definition(const QString &) const>(11, 2, QMC::AccessPublic, 0x80000000 | 5, {{
            { QMetaType::QString, 12 },
        }}),
        // Method 'definitionsForMimeType'
        QtMocHelpers::MethodData<QList<KSyntaxHighlighting::Definition>(const QString &) const>(13, 2, QMC::AccessPublic, 0x80000000 | 10, {{
            { QMetaType::QString, 12 },
        }}),
        // Method 'definitions'
        QtMocHelpers::MethodData<QList<KSyntaxHighlighting::Definition>() const>(14, 2, QMC::AccessPublic, 0x80000000 | 10),
        // Method 'themes'
        QtMocHelpers::MethodData<QList<KSyntaxHighlighting::Theme>() const>(15, 2, QMC::AccessPublic, 0x80000000 | 16),
        // Method 'theme'
        QtMocHelpers::MethodData<KSyntaxHighlighting::Theme(const QString &) const>(17, 2, QMC::AccessPublic, 0x80000000 | 18, {{
            { QMetaType::QString, 19 },
        }}),
        // Method 'defaultTheme'
        QtMocHelpers::MethodData<KSyntaxHighlighting::Theme(KSyntaxHighlighting::Repository::DefaultTheme) const>(20, 2, QMC::AccessPublic, 0x80000000 | 18, {{
            { 0x80000000 | 21, 22 },
        }}),
        // Method 'defaultTheme'
        QtMocHelpers::MethodData<KSyntaxHighlighting::Theme() const>(20, 2, QMC::AccessPublic | QMC::MethodCloned, 0x80000000 | 18),
    };
    QtMocHelpers::UintData qt_properties {
        // property 'definitions'
        QtMocHelpers::PropertyData<QList<KSyntaxHighlighting::Definition>>(14, 0x80000000 | 10, QMC::DefaultPropertyFlags | QMC::EnumOrFlag, 1),
        // property 'themes'
        QtMocHelpers::PropertyData<QList<KSyntaxHighlighting::Theme>>(15, 0x80000000 | 16, QMC::DefaultPropertyFlags | QMC::EnumOrFlag, 1),
    };
    QtMocHelpers::UintData qt_enums {
        // enum 'DefaultTheme'
        QtMocHelpers::EnumData<enum DefaultTheme>(23, 23, QMC::EnumFlags{}).add({
            {   24, DefaultTheme::LightTheme },
            {   25, DefaultTheme::DarkTheme },
        }),
    };
    return QtMocHelpers::metaObjectData<Repository, qt_meta_tag_ZN19KSyntaxHighlighting10RepositoryE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject KSyntaxHighlighting::Repository::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19KSyntaxHighlighting10RepositoryE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19KSyntaxHighlighting10RepositoryE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN19KSyntaxHighlighting10RepositoryE_t>.metaTypes,
    nullptr
} };

void KSyntaxHighlighting::Repository::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<Repository *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->aboutToReload(); break;
        case 1: _t->reloaded(); break;
        case 2: { KSyntaxHighlighting::Definition _r = _t->definitionForName((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])));
            if (_a[0]) *reinterpret_cast<KSyntaxHighlighting::Definition*>(_a[0]) = std::move(_r); }  break;
        case 3: { KSyntaxHighlighting::Definition _r = _t->definitionForFileName((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])));
            if (_a[0]) *reinterpret_cast<KSyntaxHighlighting::Definition*>(_a[0]) = std::move(_r); }  break;
        case 4: { QList<KSyntaxHighlighting::Definition> _r = _t->definitionsForFileName((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])));
            if (_a[0]) *reinterpret_cast<QList<KSyntaxHighlighting::Definition>*>(_a[0]) = std::move(_r); }  break;
        case 5: { KSyntaxHighlighting::Definition _r = _t->definitionForMimeType((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])));
            if (_a[0]) *reinterpret_cast<KSyntaxHighlighting::Definition*>(_a[0]) = std::move(_r); }  break;
        case 6: { QList<KSyntaxHighlighting::Definition> _r = _t->definitionsForMimeType((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])));
            if (_a[0]) *reinterpret_cast<QList<KSyntaxHighlighting::Definition>*>(_a[0]) = std::move(_r); }  break;
        case 7: { QList<KSyntaxHighlighting::Definition> _r = _t->definitions();
            if (_a[0]) *reinterpret_cast<QList<KSyntaxHighlighting::Definition>*>(_a[0]) = std::move(_r); }  break;
        case 8: { QList<KSyntaxHighlighting::Theme> _r = _t->themes();
            if (_a[0]) *reinterpret_cast<QList<KSyntaxHighlighting::Theme>*>(_a[0]) = std::move(_r); }  break;
        case 9: { KSyntaxHighlighting::Theme _r = _t->theme((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])));
            if (_a[0]) *reinterpret_cast<KSyntaxHighlighting::Theme*>(_a[0]) = std::move(_r); }  break;
        case 10: { KSyntaxHighlighting::Theme _r = _t->defaultTheme((*reinterpret_cast<std::add_pointer_t<KSyntaxHighlighting::Repository::DefaultTheme>>(_a[1])));
            if (_a[0]) *reinterpret_cast<KSyntaxHighlighting::Theme*>(_a[0]) = std::move(_r); }  break;
        case 11: { KSyntaxHighlighting::Theme _r = _t->defaultTheme();
            if (_a[0]) *reinterpret_cast<KSyntaxHighlighting::Theme*>(_a[0]) = std::move(_r); }  break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (Repository::*)()>(_a, &Repository::aboutToReload, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (Repository::*)()>(_a, &Repository::reloaded, 1))
            return;
    }
    if (_c == QMetaObject::ReadProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: *reinterpret_cast<QList<KSyntaxHighlighting::Definition>*>(_v) = _t->definitions(); break;
        case 1: *reinterpret_cast<QList<KSyntaxHighlighting::Theme>*>(_v) = _t->themes(); break;
        default: break;
        }
    }
}

const QMetaObject *KSyntaxHighlighting::Repository::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *KSyntaxHighlighting::Repository::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN19KSyntaxHighlighting10RepositoryE_t>.strings))
        return static_cast<void*>(this);
    return QObject::qt_metacast(_clname);
}

int KSyntaxHighlighting::Repository::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 12)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 12;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 12)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 12;
    }
    if (_c == QMetaObject::ReadProperty || _c == QMetaObject::WriteProperty
            || _c == QMetaObject::ResetProperty || _c == QMetaObject::BindableProperty
            || _c == QMetaObject::RegisterPropertyMetaType) {
        qt_static_metacall(this, _c, _id, _a);
        _id -= 2;
    }
    return _id;
}

// SIGNAL 0
void KSyntaxHighlighting::Repository::aboutToReload()
{
    QMetaObject::activate(this, &staticMetaObject, 0, nullptr);
}

// SIGNAL 1
void KSyntaxHighlighting::Repository::reloaded()
{
    QMetaObject::activate(this, &staticMetaObject, 1, nullptr);
}
QT_WARNING_POP
