/****************************************************************************
** Generated QML type registration code
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include <QtQml/qqml.h>
#include <QtQml/qqmlmoduleregistration.h>

#if __has_include(<kquicksyntaxhighlighter.h>)
#  include <kquicksyntaxhighlighter.h>
#endif
#if __has_include(<types.h>)
#  include <types.h>
#endif


#if !defined(QT_STATIC)
#define Q_QMLTYPE_EXPORT Q_DECL_EXPORT
#else
#define Q_QMLTYPE_EXPORT
#endif
Q_QMLTYPE_EXPORT void qml_register_types_org_kde_syntaxhighlighting()
{
    QT_WARNING_PUSH QT_WARNING_DISABLE_DEPRECATED
    qmlRegisterTypesAndRevisions<DefinitionForeign>("org.kde.syntaxhighlighting", 1);
    qmlRegisterTypesAndRevisions<KQuickSyntaxHighlighter>("org.kde.syntaxhighlighting", 1);
    qmlRegisterTypesAndRevisions<RepositoryForeign>("org.kde.syntaxhighlighting", 1);
    qmlRegisterEnum<KSyntaxHighlighting::Repository::DefaultTheme>("KSyntaxHighlighting::Repository::DefaultTheme");
    qmlRegisterTypesAndRevisions<ThemeForeign>("org.kde.syntaxhighlighting", 1);
    qmlRegisterEnum<KSyntaxHighlighting::Theme::TextStyle>("KSyntaxHighlighting::Theme::TextStyle");
    qmlRegisterEnum<KSyntaxHighlighting::Theme::EditorColorRole>("KSyntaxHighlighting::Theme::EditorColorRole");
    QT_WARNING_POP
    qmlRegisterModule("org.kde.syntaxhighlighting", 1, 0);
}

static const QQmlModuleRegistration orgkdesyntaxhighlightingRegistration("org.kde.syntaxhighlighting", qml_register_types_org_kde_syntaxhighlighting);
